"""Declarations for Python renderings of basic JavaScript types."""

from __future__ import annotations

from collections.abc import (
    Awaitable,
    Callable,
    Generator,
    MutableMapping,
    MutableSequence,
)
from datetime import datetime
from typing import Any, TypeAlias


class JSUndefinedType:
    """The JavaScript undefined type.

    Where JavaScript null is represented as None, undefined is represented as this
    type."""

    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return "JSUndefined"


JSUndefined = JSUndefinedType()


class JSObject:
    """A JavaScript object."""


class JSMappedObject(
    MutableMapping["PythonJSConvertedTypes", "PythonJSConvertedTypes"], JSObject
):
    """A JavaScript object with Pythonic MutableMapping methods (`keys()`,
    `__getitem__()`, etc).

    `keys()` and `__iter__()` will return properties from any prototypes as well as this
    object, as if using a for-in statement in JavaScript.
    """


class JSArray(MutableSequence["PythonJSConvertedTypes"], JSObject):
    """JavaScript array.

    Has Pythonic MutableSequence methods (e.g., `insert()`, `__getitem__()`, ...).
    """


class JSFunction(JSMappedObject):
    """JavaScript function.

    This type is returned by synchronous MiniRacer contexts.

    You can call this object from Python, passing in positional args to match what the
    JavaScript function expects.

    In synchronous code you may supply an additional keyword argument, `timeout_sec`.

    If you are running in an async context in the same event loop as Miniracer, you must
    not supply a non-None value for timeout_sec. Instead call func.cancelable() to
    obtain a CancellableJSFunction, and use a construct like asyncio.wait_for(...) to
    apply a timeout.
    """

    def __call__(
        self,
        *args: PythonJSConvertedTypes,
        this: JSObject | JSUndefinedType = JSUndefined,
        timeout_sec: float | None = None,
    ) -> PythonJSConvertedTypes:
        raise NotImplementedError

    def cancelable(self) -> CancelableJSFunction:
        raise NotImplementedError


class CancelableJSFunction(JSMappedObject):
    """JavaScript function.

    This type is returned by JSFunction.cancelable().

    You can call this object from Python, passing in positional args to match what the
    JavaScript function expects. Calls on the Python side are async, regardless of
    whether the underlying JS function is async (so a call to an async JS function will
    return a JSPromise, requiring a "double await" in Python to get the result).
    """

    async def __call__(
        self,
        *args: PythonJSConvertedTypes,
        this: JSObject | JSUndefinedType = JSUndefined,
    ) -> PythonJSConvertedTypes:
        raise NotImplementedError


class JSSymbol(JSMappedObject):
    """JavaScript symbol."""


class JSPromise(JSObject):
    """JavaScript Promise.

    To get a value, call `promise.get()` (which blocks) or `await promise` (in async
    code). Both operations will raise a Python exception if the JavaScript Promise is
    rejected.
    """

    def get(self, *, timeout: float | None = None) -> PythonJSConvertedTypes:
        raise NotImplementedError

    def __await__(self) -> Generator[Any, None, Any]:
        raise NotImplementedError


PythonJSConvertedTypes: TypeAlias = (
    JSUndefinedType
    | bool
    | int
    | float
    | str
    | JSObject
    | datetime
    | memoryview
    | JSPromise
    | JSFunction
    | CancelableJSFunction
    | JSMappedObject
    | JSSymbol
    | JSArray
    | None
)


PyJsFunctionType = Callable[..., Awaitable[PythonJSConvertedTypes]]



from __future__ import annotations

import asyncio
import queue
from contextlib import asynccontextmanager, contextmanager, suppress
from dataclasses import dataclass, field
from traceback import format_exc
from typing import TYPE_CHECKING, Any, NewType, Protocol, cast

from py_mini_racer._dll import init_mini_racer
from py_mini_racer._exc import JSEvalException, JSPromiseError
from py_mini_racer._types import (
    CancelableJSFunction,
    JSArray,
    JSFunction,
    JSMappedObject,
    JSObject,
    JSPromise,
    JSUndefined,
    JSUndefinedType,
    PyJsFunctionType,
    PythonJSConvertedTypes,
)
from py_mini_racer._value_handle import ValueHandle

if TYPE_CHECKING:
    import ctypes
    from collections.abc import AsyncGenerator, Callable, Coroutine, Generator, Iterator

    from py_mini_racer._value_handle import RawValueHandleType


def context_count() -> int:
    """For tests only: how many context handles are still allocated?"""

    dll = init_mini_racer(ignore_duplicate_init=True)
    return int(dll.mr_context_count())


ContextType = NewType("ContextType", object)
_UNCANCELABLE_TASK_CALLBACK_ID = 0


@dataclass(frozen=True)
class _TaskSet:
    """This is a very very simplistic standin for Python 3.11+ TaskGroups (whereas we
    are still targeting Python 3.10)."""

    _event_loop: asyncio.AbstractEventLoop
    _ongoing_tasks: set[asyncio.Task[PythonJSConvertedTypes]]

    def start_task(self, coro: Coroutine[Any, Any, None]) -> None:
        task = self._event_loop.create_task(coro)
        self._ongoing_tasks.add(task)
        task.add_done_callback(self._ongoing_tasks.discard)


@asynccontextmanager
async def _make_task_set(
    event_loop: asyncio.AbstractEventLoop,
) -> AsyncGenerator[_TaskSet, None]:
    ongoing_tasks: set[asyncio.Task[PythonJSConvertedTypes]] = set()

    try:
        yield _TaskSet(event_loop, ongoing_tasks)
    finally:
        for t in list(ongoing_tasks):
            with suppress(asyncio.CancelledError):
                t.cancel()
                await t


class ObjectFactory(Protocol):
    def value_handle_to_python(
        self, ctx: Context, val_handle: ValueHandle
    ) -> PythonJSConvertedTypes: ...

    def python_to_value_handle(
        self, ctx: Context, obj: PythonJSConvertedTypes
    ) -> ValueHandle: ...


def get_running_loop_or_none() -> asyncio.AbstractEventLoop | None:
    try:
        return asyncio.get_running_loop()
    except RuntimeError:
        return None


@dataclass(frozen=True)
class Context:
    """Wrapper for all operations involving the DLL and C++ MiniRacer::Context."""

    _dll: ctypes.CDLL
    _ctx: ContextType
    event_loop: asyncio.AbstractEventLoop
    _object_factory: ObjectFactory
    _next_async_callback_id: Iterator[int]
    _active_cancelable_mr_task_callbacks: dict[int, Callable[[ValueHandle], None]] = (
        field(default_factory=dict)
    )
    _non_cancelable_mr_task_results_queue: queue.Queue[ValueHandle] = field(
        default_factory=queue.Queue
    )

    def v8_version(self) -> str:
        return str(self._dll.mr_v8_version().decode("utf-8"))

    def v8_is_using_sandbox(self) -> bool:
        """Checks for enablement of the V8 Sandbox. See https://v8.dev/blog/sandbox."""

        return bool(self._dll.mr_v8_is_using_sandbox())

    def handle_callback_from_v8(
        self, callback_id: int, raw_val_handle: RawValueHandleType
    ) -> None:
        # Handle a callback from within the v8::Isolate.
        # All work on the Isolate is blocked until this callback returns. That may
        # may in turn be blocking incoming calls from Python, including other threads,
        # asyncio event loops, etc. So we need to get out fast!
        # We limit ourselves to wrapping the incoming handle (so we don't leak memory)
        # and enqueing the incoming work for decoupled processing.

        val_handle = self._wrap_raw_handle(raw_val_handle)

        if callback_id == _UNCANCELABLE_TASK_CALLBACK_ID:
            self._non_cancelable_mr_task_results_queue.put(val_handle)
        else:
            self.event_loop.call_soon_threadsafe(
                self._handle_callback_from_v8_on_event_loop, callback_id, val_handle
            )

    def _handle_callback_from_v8_on_event_loop(
        self, callback_id: int, val_handle: ValueHandle
    ) -> None:
        try:
            callback = self._active_cancelable_mr_task_callbacks[callback_id]
        except KeyError:
            # Assume this callback was intentionally cancelled:
            return

        callback(val_handle)

    @contextmanager
    def _register_cancelable_mr_task_callback(
        self, func: Callable[[ValueHandle], None]
    ) -> Generator[int, None, None]:
        callback_id = next(self._next_async_callback_id)

        self._active_cancelable_mr_task_callbacks[callback_id] = func

        try:
            yield callback_id
        finally:
            self._active_cancelable_mr_task_callbacks.pop(callback_id)

    async def eval_cancelable(self, code: str) -> PythonJSConvertedTypes:
        code_handle = self._python_to_value_handle(code)

        return await self._run_cancelable_mr_task(self._dll.mr_eval, code_handle.raw)

    def eval(self, code: str) -> PythonJSConvertedTypes:
        code_handle = self._python_to_value_handle(code)

        return self._run_uncancelable_mr_task(self._dll.mr_eval, code_handle.raw)

    async def await_promise(self, promise: JSPromise) -> PythonJSConvertedTypes:
        promise_handle = self._python_to_value_handle(promise)
        then_name_handle = self._python_to_value_handle("then")

        then_func = cast(
            "JSFunction",
            self._value_handle_to_python(
                self._wrap_raw_handle(
                    self._dll.mr_get_object_item(
                        self._ctx, promise_handle.raw, then_name_handle.raw
                    )
                )
            ),
        )

        future: asyncio.Future[PythonJSConvertedTypes] = self.event_loop.create_future()

        def on_resolved(val_handle: ValueHandle) -> None:
            if future.cancelled():
                return

            future.set_result(
                cast("JSArray", self._value_handle_to_python(val_handle))[0]
            )

        def on_rejected(val_handle: ValueHandle) -> None:
            if future.cancelled():
                return

            value = cast("JSArray", self._value_handle_to_python(val_handle))[0]
            if not isinstance(value, JSMappedObject):
                msg = str(value)
            elif "stack" in value:
                msg = cast("str", value["stack"])
            else:
                msg = str(value)

            future.set_exception(JSPromiseError(msg))

        with (
            self._register_js_notification(on_resolved) as on_resolved_js_func,
            self._register_js_notification(on_rejected) as on_rejected_js_func,
        ):
            then_func(on_resolved_js_func, on_rejected_js_func, this=promise)

            return await future

    def get_identity_hash(self, obj: JSObject) -> int:
        obj_handle = self._python_to_value_handle(obj)

        return cast(
            "int",
            self._value_handle_to_python(
                self._wrap_raw_handle(
                    self._dll.mr_get_identity_hash(self._ctx, obj_handle.raw)
                )
            ),
        )

    def get_own_property_names(
        self, obj: JSObject
    ) -> tuple[PythonJSConvertedTypes, ...]:
        obj_handle = self._python_to_value_handle(obj)

        names = self._value_handle_to_python(
            self._wrap_raw_handle(
                self._dll.mr_get_own_property_names(self._ctx, obj_handle.raw)
            )
        )
        if not isinstance(names, JSArray):
            raise TypeError
        return tuple(names)

    def get_object_item(
        self, obj: JSObject, key: PythonJSConvertedTypes
    ) -> PythonJSConvertedTypes:
        obj_handle = self._python_to_value_handle(obj)
        key_handle = self._python_to_value_handle(key)

        return self._value_handle_to_python(
            self._wrap_raw_handle(
                self._dll.mr_get_object_item(self._ctx, obj_handle.raw, key_handle.raw)
            )
        )

    def set_object_item(
        self, obj: JSObject, key: PythonJSConvertedTypes, val: PythonJSConvertedTypes
    ) -> None:
        obj_handle = self._python_to_value_handle(obj)
        key_handle = self._python_to_value_handle(key)
        val_handle = self._python_to_value_handle(val)

        # Convert the value just to convert any exceptions (and GC the result)
        self._value_handle_to_python(
            self._wrap_raw_handle(
                self._dll.mr_set_object_item(
                    self._ctx, obj_handle.raw, key_handle.raw, val_handle.raw
                )
            )
        )

    def del_object_item(self, obj: JSObject, key: PythonJSConvertedTypes) -> None:
        obj_handle = self._python_to_value_handle(obj)
        key_handle = self._python_to_value_handle(key)

        # Convert the value just to convert any exceptions (and GC the result)
        self._value_handle_to_python(
            self._wrap_raw_handle(
                self._dll.mr_del_object_item(self._ctx, obj_handle.raw, key_handle.raw)
            )
        )

    def del_from_array(self, arr: JSArray, index: int) -> None:
        arr_handle = self._python_to_value_handle(arr)

        # Convert the value just to convert any exceptions (and GC the result)
        self._value_handle_to_python(
            self._wrap_raw_handle(
                self._dll.mr_splice_array(self._ctx, arr_handle.raw, index, 1, None)
            )
        )

    def array_insert(
        self, arr: JSArray, index: int, new_val: PythonJSConvertedTypes
    ) -> None:
        arr_handle = self._python_to_value_handle(arr)
        new_val_handle = self._python_to_value_handle(new_val)

        # Convert the value just to convert any exceptions (and GC the result)
        self._value_handle_to_python(
            self._wrap_raw_handle(
                self._dll.mr_splice_array(
                    self._ctx, arr_handle.raw, index, 0, new_val_handle.raw
                )
            )
        )

    def array_push(self, arr: JSArray, new_val: PythonJSConvertedTypes) -> None:
        arr_handle = self._python_to_value_handle(arr)
        new_val_handle = self._python_to_value_handle(new_val)

        # Convert the value just to convert any exceptions (and GC the result)
        self._value_handle_to_python(
            self._wrap_raw_handle(
                self._dll.mr_array_push(self._ctx, arr_handle.raw, new_val_handle.raw)
            )
        )

    def are_we_running_on_the_mini_racer_event_loop(self) -> bool:
        return get_running_loop_or_none() is self.event_loop

    async def call_function_cancelable(
        self,
        func: CancelableJSFunction | JSFunction,
        *args: PythonJSConvertedTypes,
        this: JSObject | JSUndefinedType = JSUndefined,
    ) -> PythonJSConvertedTypes:
        argv = cast("JSArray", self.eval("[]"))
        for arg in args:
            argv.append(arg)

        func_handle = self._python_to_value_handle(func)
        this_handle = self._python_to_value_handle(this)
        argv_handle = self._python_to_value_handle(argv)

        return await self._run_cancelable_mr_task(
            self._dll.mr_call_function,
            func_handle.raw,
            this_handle.raw,
            argv_handle.raw,
        )

    def call_function(
        self,
        func: JSFunction,
        *args: PythonJSConvertedTypes,
        this: JSObject | JSUndefinedType = JSUndefined,
    ) -> PythonJSConvertedTypes:
        argv = cast("JSArray", self.eval("[]"))
        for arg in args:
            argv.append(arg)

        func_handle = self._python_to_value_handle(func)
        this_handle = self._python_to_value_handle(this)
        argv_handle = self._python_to_value_handle(argv)

        return self._run_uncancelable_mr_task(
            self._dll.mr_call_function,
            func_handle.raw,
            this_handle.raw,
            argv_handle.raw,
        )

    def set_hard_memory_limit(self, limit: int) -> None:
        self._dll.mr_set_hard_memory_limit(self._ctx, limit)

    def set_soft_memory_limit(self, limit: int) -> None:
        self._dll.mr_set_soft_memory_limit(self._ctx, limit)

    def was_hard_memory_limit_reached(self) -> bool:
        return bool(self._dll.mr_hard_memory_limit_reached(self._ctx))

    def was_soft_memory_limit_reached(self) -> bool:
        return bool(self._dll.mr_soft_memory_limit_reached(self._ctx))

    def low_memory_notification(self) -> None:
        self._dll.mr_low_memory_notification(self._ctx)

    def heap_stats(self) -> str:
        return cast(
            "str",
            self._value_handle_to_python(
                self._wrap_raw_handle(self._dll.mr_heap_stats(self._ctx))
            ),
        )

    def heap_snapshot(self) -> str:
        """Return a snapshot of the V8 isolate heap."""

        return cast(
            "str",
            self._value_handle_to_python(
                self._wrap_raw_handle(self._dll.mr_heap_snapshot(self._ctx))
            ),
        )

    def value_count(self) -> int:
        """For tests only: how many value handles are still allocated?"""

        return int(self._dll.mr_value_count(self._ctx))

    @contextmanager
    def _register_js_notification(
        self, func: Callable[[ValueHandle], None]
    ) -> Generator[JSFunction, None, None]:
        """Create a "notification": an async, one-way callback function, from JavaScript
        to Python.

        "One-way" here means the function returns nothing. "async" means that on the JS
        side, the function returns before it has been processed on the Python side."""

        with self._register_cancelable_mr_task_callback(func) as callback_id:
            yield cast(
                "JSFunction",
                self._value_handle_to_python(
                    self._wrap_raw_handle(
                        self._dll.mr_make_js_callback(self._ctx, callback_id)
                    )
                ),
            )

    @asynccontextmanager
    async def wrap_py_function_as_js_function(
        self, func: PyJsFunctionType
    ) -> AsyncGenerator[JSFunction, None]:
        async def await_into_js_promise_resolvers(val_handle: ValueHandle) -> None:
            params = self._value_handle_to_python(val_handle)
            arguments, resolve, reject = cast("JSArray", params)
            try:
                result = await func(*cast("JSArray", arguments))
                cast("JSFunction", resolve)(result)
            except Exception:  # noqa: BLE001
                # Convert this Python exception into a JS exception so we can send
                # it into JS:
                err_maker = cast("JSFunction", self.eval("s => new Error(s)"))
                cast("JSFunction", reject)(
                    err_maker(f"Error running Python function:\n{format_exc()}")
                )

        async with _make_task_set(self.event_loop) as task_set:
            with self._register_js_notification(
                lambda val_handle: task_set.start_task(
                    await_into_js_promise_resolvers(val_handle)
                )
            ) as js_to_py_notification:
                # Every time our callback is called from JS, on the JS side we
                # instantiate a JS Promise and immediately pass its resolution functions
                # into our Python callback function. While we wait on Python's asyncio
                # loop to process this call, we can return the Promise to the JS caller,
                # thus exposing what looks like an ordinary async function on the JS
                # side of things.
                wrap_outbound_calls_with_js_promises = cast(
                    "JSFunction",
                    self.eval(
                        """
fn => {
    return (...arguments) => {
        let p = Promise.withResolvers();

        fn(arguments, p.resolve, p.reject);

        return p.promise;
    }
}
"""
                    ),
                )

                yield cast(
                    "JSFunction",
                    wrap_outbound_calls_with_js_promises(js_to_py_notification),
                )

    def _wrap_raw_handle(self, raw: RawValueHandleType) -> ValueHandle:
        return ValueHandle(lambda: self._free(raw), raw)

    def create_intish_val(self, val: int, typ: int) -> ValueHandle:
        return self._wrap_raw_handle(self._dll.mr_alloc_int_val(self._ctx, val, typ))

    def create_doublish_val(self, val: float, typ: int) -> ValueHandle:
        return self._wrap_raw_handle(self._dll.mr_alloc_double_val(self._ctx, val, typ))

    def create_string_val(self, val: str, typ: int) -> ValueHandle:
        b = val.encode("utf-8")
        return self._wrap_raw_handle(
            self._dll.mr_alloc_string_val(self._ctx, b, len(b), typ)
        )

    def _free(self, raw: RawValueHandleType) -> None:
        self._dll.mr_free_value(self._ctx, raw)

    async def _run_cancelable_mr_task(
        self,
        dll_method: Any,  # noqa: ANN401
        *args: Any,  # noqa: ANN401
    ) -> PythonJSConvertedTypes:
        """Manages cancelable tasks within the MiniRacer DLL.

        Several MiniRacer functions (JS evaluation and 2 heap stats calls) are
        cancelable and asynchronous. They take a function callback and callback data
        parameter, and they return a task handle.

        In this method, we create a future for each callback to get the right data to
        the right caller, and we manage the lifecycle of the task and task handle.
        """

        future: asyncio.Future[PythonJSConvertedTypes] = asyncio.Future()

        def callback(val_handle: ValueHandle) -> None:
            if future.cancelled():
                return

            try:
                value = self._value_handle_to_python(val_handle)
            except JSEvalException as e:
                future.set_exception(e)
                return

            future.set_result(value)

        with self._register_cancelable_mr_task_callback(callback) as callback_id:
            # Start the task:
            task_id = dll_method(self._ctx, *args, callback_id)
            try:
                return await future
            finally:
                # Cancel the task if it's not already done (this call is ignored if it's
                # already done)
                self._dll.mr_cancel_task(self._ctx, task_id)

    def _run_uncancelable_mr_task(
        self,
        dll_method: Any,  # noqa: ANN401
        *args: Any,  # noqa: ANN401
    ) -> PythonJSConvertedTypes:
        """Like _run_cancelable_mr_task, but eschewing cancellation semantics and
        instead just waiting on a result synchronously."""

        # self._non_cancelable_mr_task_results_queue is single file, with no
        # higher-level ordering mechanism, so it's important that we use it only from
        # the event loop thread, to keep things in order:
        assert self.are_we_running_on_the_mini_racer_event_loop()

        _task_id = dll_method(self._ctx, *args, _UNCANCELABLE_TASK_CALLBACK_ID)
        val_handle = self._non_cancelable_mr_task_results_queue.get()
        return self._value_handle_to_python(val_handle)

    def _value_handle_to_python(
        self, val_handle: ValueHandle
    ) -> PythonJSConvertedTypes:
        return self._object_factory.value_handle_to_python(self, val_handle)

    def _python_to_value_handle(self, obj: PythonJSConvertedTypes) -> ValueHandle:
        return self._object_factory.python_to_value_handle(self, obj)



from __future__ import annotations

import asyncio
import json
from contextlib import (
    AbstractContextManager,
    asynccontextmanager,
    contextmanager,
    suppress,
)
from itertools import count
from json import JSONEncoder
from threading import Thread
from typing import TYPE_CHECKING, Any, ClassVar

from py_mini_racer._context import Context, ContextType, get_running_loop_or_none
from py_mini_racer._dll import init_mini_racer, mr_callback_func
from py_mini_racer._exc import JSTimeoutException, WrongReturnTypeException
from py_mini_racer._objects import ObjectFactoryImpl
from py_mini_racer._set_timeout import INSTALL_SET_TIMEOUT

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Generator
    from types import TracebackType

    from typing_extensions import Self

    from py_mini_racer._types import (
        JSFunction,
        PyJsFunctionType,
        PythonJSConvertedTypes,
    )
    from py_mini_racer._value_handle import RawValueHandleType


class MiniRacer:
    """
    MiniRacer evaluates JavaScript code using a V8 isolate.

    A MiniRacer instance can be explicitly closed using the close() method, or by using
    the MiniRacer as a context manager, i.e,:

    with MiniRacer() as mr:
        ...

    The MiniRacer instance will otherwise clean up the underlying V8 resources upon
    garbage collection.

    Attributes:
        json_impl: JSON module used by helper methods default is
            [json](https://docs.python.org/3/library/json.html)
    """

    json_impl: ClassVar[Any] = json

    def __init__(self, context: Context | None = None) -> None:
        if context is None:
            self._own_context_maker: AbstractContextManager[Context] | None = (
                _make_context()
            )
            self._ctx: Context | None = self._own_context_maker.__enter__()
        else:
            self._own_context_maker = None
            self._ctx = context

        self.eval(INSTALL_SET_TIMEOUT)

    def close(
        self,
        exc_type: type[BaseException] | None = None,
        exc_val: BaseException | None = None,
        exc_tb: TracebackType | None = None,
    ) -> None:
        """Close this MiniRacer instance.

        It is an error to use this MiniRacer instance or any JS objects returned by it
        after calling this method.
        """
        own_context_maker = self._own_context_maker
        self._own_context_maker = None
        self._ctx = None

        if own_context_maker is not None:
            own_context_maker.__exit__(exc_type, exc_val, exc_tb)

    def __del__(self) -> None:
        # Ignore ordering problems on process teardown.
        # (A user who wants consistent teardown should use `with MiniRacer() as ctx`
        # which makes the cleanup deterministic.)
        with suppress(Exception):
            self.close()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.close(exc_type, exc_val, exc_tb)

    @property
    def v8_version(self) -> str:
        """Return the V8 version string."""
        assert self._ctx is not None
        return self._ctx.v8_version()

    def eval(
        self,
        code: str,
        timeout: float | None = None,
        timeout_sec: float | None = None,
        max_memory: int | None = None,
    ) -> PythonJSConvertedTypes:
        """Evaluate JavaScript code in the V8 isolate.

        Side effects from the JavaScript evaluation is persisted inside a context
        (meaning variables set are kept for the next evaluation).

        The JavaScript value returned by the last expression in `code` is converted to
        a Python value and returned by this method. Only primitive types are supported
        (numbers, strings, buffers...). Use the
        [py_mini_racer.MiniRacer.execute][] method to return more complex
        types such as arrays or objects.

        The evaluation can be interrupted by an exception for several reasons: a limit
        was reached, the code could not be parsed, a returned value could not be
        converted to a Python value.

        Args:
            code: JavaScript code
            timeout: number of milliseconds after which the execution is interrupted.
                This is deprecated; use timeout_sec instead.
            timeout_sec: number of seconds after which the execution is interrupted
            max_memory: hard memory limit, in bytes, after which the execution is
                interrupted.
        """

        if max_memory is not None:
            self.set_hard_memory_limit(max_memory)

        if timeout:
            # PyMiniRacer unfortunately uses milliseconds while Python and
            # Système international d'unités use seconds.
            timeout_sec = timeout / 1000

        ctx = self._ctx
        assert ctx is not None

        if not ctx.are_we_running_on_the_mini_racer_event_loop():

            async def run() -> PythonJSConvertedTypes:
                try:
                    return await asyncio.wait_for(
                        ctx.eval_cancelable(code), timeout=timeout_sec
                    )
                except asyncio.TimeoutError as e:
                    raise JSTimeoutException from e

            return asyncio.run_coroutine_threadsafe(run(), ctx.event_loop).result()

        assert timeout_sec is None, (
            "To apply a timeout in an async context, use "
            "`await asyncio.wait_for(mr.eval_cancelable(your_params), "
            "timeout=your_timeout)`"
        )

        return ctx.eval(code)

    async def eval_cancelable(self, code: str) -> PythonJSConvertedTypes:
        """Evaluate JavaScript code in the V8 isolate.

        Similar to eval(), but runaway calls can be canceled by canceling the
        coroutine's task, e.g., using:

            await asyncio.wait_for(mr.eval_cancelable(...), timeout=some_timeout)

        """

        assert self._ctx is not None

        return await self._ctx.eval_cancelable(code)

    def execute(
        self,
        expr: str,
        timeout: float | None = None,
        timeout_sec: float | None = None,
        max_memory: int | None = None,
    ) -> Any:  # noqa: ANN401
        """Helper to evaluate a JavaScript expression and return composite types.

        Returned value is serialized to JSON inside the V8 isolate and deserialized
        using `json_impl`.

        Args:
            expr: JavaScript expression
            timeout: number of milliseconds after which the execution is interrupted.
                This is deprecated; use timeout_sec instead.
            timeout_sec: number of seconds after which the execution is interrupted
            max_memory: hard memory limit, in bytes, after which the execution is
                interrupted.
        """

        if timeout:
            # PyMiniRacer unfortunately uses milliseconds while Python and
            # Système international d'unités use seconds.
            timeout_sec = timeout / 1000

        wrapped_expr = f"JSON.stringify((function(){{return ({expr})}})())"
        ret = self.eval(wrapped_expr, timeout_sec=timeout_sec, max_memory=max_memory)
        if not isinstance(ret, str):
            raise WrongReturnTypeException(type(ret))
        return self.json_impl.loads(ret)

    def call(
        self,
        expr: str,
        *args: Any,  # noqa: ANN401
        encoder: type[JSONEncoder] | None = None,
        timeout: float | None = None,
        timeout_sec: float | None = None,
        max_memory: int | None = None,
    ) -> Any:  # noqa: ANN401
        """Helper to call a JavaScript function and return compositve types.

        The `expr` argument refers to a JavaScript function in the current V8
        isolate context. Further positional arguments are serialized using the JSON
        implementation `json_impl` and passed to the JavaScript function as arguments.

        Returned value is serialized to JSON inside the V8 isolate and deserialized
        using `json_impl`.

        Args:
            expr: JavaScript expression referring to a function
            encoder: Custom JSON encoder
            timeout: number of milliseconds after which the execution is
                interrupted.
            timeout_sec: number of seconds after which the execution is interrupted
            max_memory: hard memory limit, in bytes, after which the execution is
                interrupted
        """

        if timeout:
            # PyMiniRacer unfortunately uses milliseconds while Python and
            # Système international d'unités use seconds.
            timeout_sec = timeout / 1000

        json_args = self.json_impl.dumps(args, separators=(",", ":"), cls=encoder)
        js = f"{expr}.apply(this, {json_args})"
        return self.execute(js, timeout_sec=timeout_sec, max_memory=max_memory)

    @asynccontextmanager
    async def wrap_py_function(
        self, func: PyJsFunctionType
    ) -> AsyncGenerator[JSFunction, None]:
        """Wrap a Python function such that it can be called from JS.

        To be wrapped and exposed in JavaScript, a Python function should:

          1. Be async,
          2. Accept variable positional arguments each of type PythonJSConvertedTypes,
             and
          3. Return one value of type PythonJSConvertedTypes (a type union which
             includes None).

        The function is rendered on the JavaScript side as an async function (i.e., a
        function which returns a Promise).

        Returns:
            An async context manager which, when entered, yields a JS Function which
            can be passed into MiniRacer and called by JS code.
        """

        assert self._ctx is not None

        async with self._ctx.wrap_py_function_as_js_function(func) as js_func:
            yield js_func

    def set_hard_memory_limit(self, limit: int) -> None:
        """Set a hard memory limit on this V8 isolate.

        JavaScript execution will be terminated when this limit is reached.

        :param int limit: memory limit in bytes or 0 to reset the limit
        """

        assert self._ctx is not None
        self._ctx.set_hard_memory_limit(limit)

    def set_soft_memory_limit(self, limit: int) -> None:
        """Set a soft memory limit on this V8 isolate.

        The Garbage Collection will use a more aggressive strategy when
        the soft limit is reached but the execution will not be stopped.

        :param int limit: memory limit in bytes or 0 to reset the limit
        """

        assert self._ctx is not None
        self._ctx.set_soft_memory_limit(limit)

    def was_hard_memory_limit_reached(self) -> bool:
        """Return true if the hard memory limit was reached on the V8 isolate."""

        assert self._ctx is not None
        return self._ctx.was_hard_memory_limit_reached()

    def was_soft_memory_limit_reached(self) -> bool:
        """Return true if the soft memory limit was reached on the V8 isolate."""

        assert self._ctx is not None
        return self._ctx.was_soft_memory_limit_reached()

    def low_memory_notification(self) -> None:
        """Ask the V8 isolate to collect memory more aggressively."""

        assert self._ctx is not None
        self._ctx.low_memory_notification()

    def heap_stats(self) -> Any:  # noqa: ANN401
        """Return the V8 isolate heap statistics."""

        assert self._ctx is not None
        return self.json_impl.loads(self._ctx.heap_stats())

    def heap_snapshot(self) -> Any:  # noqa: ANN401
        """Return a snapshot of the V8 isolate heap."""

        assert self._ctx is not None
        return self.json_impl.loads(self._ctx.heap_snapshot())


@contextmanager
def _running_event_loop(
    event_loop: asyncio.AbstractEventLoop | None = None,
) -> Generator[asyncio.AbstractEventLoop, None, None]:
    """Pick an asyncio loop. In descending order of precedence:

    1. The caller-specified one,
    2. The running one (defined if we're being called from async context), or
    3. One we create and launch a thread for, on the spot.
    """

    event_loop = event_loop or get_running_loop_or_none()

    if event_loop is not None:
        yield event_loop
        return

    event_loop = asyncio.new_event_loop()

    def run_event_loop() -> None:
        asyncio.set_event_loop(event_loop)
        assert event_loop is not None
        event_loop.run_forever()
        event_loop.close()

    event_loop_thread = Thread(target=run_event_loop, daemon=True)
    event_loop_thread.start()

    try:
        yield event_loop
    finally:
        event_loop.call_soon_threadsafe(event_loop.stop)
        event_loop_thread.join()


@contextmanager
def _make_context(
    event_loop: asyncio.AbstractEventLoop | None = None,
) -> Generator[Context, None, None]:
    dll = init_mini_racer(ignore_duplicate_init=True)

    context: Context

    # define an all-purpose callback:
    @mr_callback_func
    def mr_callback(callback_id: int, raw_val_handle: RawValueHandleType) -> None:
        nonlocal context
        context.handle_callback_from_v8(callback_id, raw_val_handle)

    next_cancelable_task_callback_id = count()
    # reserve 0 as the callback for tasks we don't bother canceling; see
    # _UNCANCELABLE_TASK_CALLBACK_ID:
    _ = next(next_cancelable_task_callback_id)

    ctx = ContextType(dll.mr_init_context(mr_callback))
    try:
        with _running_event_loop(event_loop) as loop:
            context = Context(
                dll, ctx, loop, ObjectFactoryImpl(), next_cancelable_task_callback_id
            )
            yield context
    finally:
        dll.mr_free_context(ctx)


@contextmanager
def mini_racer(
    event_loop: asyncio.AbstractEventLoop | None = None,
) -> Generator[MiniRacer, None]:
    with _make_context(event_loop) as ctx:
        yield MiniRacer(ctx)


# Compatibility with versions 0.4 & 0.5
StrictMiniRacer = MiniRacer



from __future__ import annotations

from py_mini_racer import MiniRacer
from tests.gc_check import assert_no_v8_objects


def test_heap_stats() -> None:
    mr = MiniRacer()

    assert mr.heap_stats()["used_heap_size"] > 0
    assert mr.heap_stats()["total_heap_size"] > 0

    assert_no_v8_objects(mr)


def test_heap_snapshot() -> None:
    mr = MiniRacer()

    assert mr.heap_snapshot()["edges"]
    assert mr.heap_snapshot()["strings"]

    assert_no_v8_objects(mr)


def test_no_handle_leak() -> None:
    mr = MiniRacer()

    big_obj_maker = "Array.from({ length: 1000000 }, (_, i) => i)"

    mr.eval(big_obj_maker)

    one_object_size = mr.heap_stats()["used_heap_size"]

    for _ in range(100):
        mr.eval(big_obj_maker)

    many_object_size = mr.heap_stats()["used_heap_size"]

    # It's okay if making a lot of big objects and disposing causes V8 to increase the
    # heap, but only to a point. Let it allocate 3x the memory for a repeated operation
    # that, in principle, uses a fixed amount of memory and freess it on every loop:
    allowable_ratio = 3
    assert many_object_size / one_object_size < allowable_ratio

    assert_no_v8_objects(mr)



"""Test Python functions exposed as JS."""

from __future__ import annotations

from asyncio import Future, gather
from asyncio import run as asyncio_run
from asyncio import sleep as asyncio_sleep
from time import time
from typing import TYPE_CHECKING, Any, NoReturn, cast

import pytest

from py_mini_racer import JSFunction, JSPromise, JSPromiseError, mini_racer
from tests.gc_check import async_assert_no_v8_objects

if TYPE_CHECKING:
    from py_mini_racer import JSFunction
    from py_mini_racer._mini_racer import MiniRacer


_NUM_LOOPS = 10


def test_basic() -> None:
    data = []

    async def append(*args: Any) -> str:  # noqa: ANN401
        data.append(args)
        return "foobar"

    async def define_and_use_function(mr: MiniRacer) -> None:
        async with mr.wrap_py_function(append) as jsfunc:
            # "Install" our JS function on the global "this" object:
            cast("JSFunction", mr.eval("x => this.func = x"))(jsfunc)

            # Call our function a couple times, through JS:
            assert await cast("JSPromise", mr.eval("this.func(42)")) == "foobar"
            assert await cast("JSPromise", mr.eval('this.func("blah")')) == "foobar"

        assert data == [(42,), ("blah",)]
        data[:] = []

    async def run() -> None:
        with mini_racer() as mr:
            for _ in range(_NUM_LOOPS):
                await define_and_use_function(mr)

            await async_assert_no_v8_objects(mr)

    for _ in range(_NUM_LOOPS):
        asyncio_run(run())


def test_exception() -> None:
    # Test a Python callback which raises exceptions

    async def append(*args: Any) -> NoReturn:  # noqa: ANN401
        del args
        boo = "boo"
        raise RuntimeError(boo)

    async def run() -> None:
        with mini_racer() as mr:
            async with mr.wrap_py_function(append) as jsfunc:
                # "Install" our JS function on the global "this" object:
                cast("JSFunction", mr.eval("x => this.func = x"))(jsfunc)

                with pytest.raises(JSPromiseError) as exc_info:
                    await cast("JSPromise", mr.eval("this.func(42)"))

                assert exc_info.value.args[0].startswith(
                    """\
JavaScript rejected promise with reason: Error: Error running Python function:
Traceback (most recent call last):
"""
                )

                assert exc_info.value.args[0].endswith(
                    """\

    at <anonymous>:1:6
"""
                )

            del exc_info, jsfunc
            await async_assert_no_v8_objects(mr)

    for _ in range(_NUM_LOOPS):
        asyncio_run(run())


def test_slow() -> None:
    # Test a Python callback which runs slowly, but is faster in parallel.
    data = []

    async def append(*args: Any) -> str:  # noqa: ANN401
        await asyncio_sleep(1)
        data.append(args)
        return "foobar"

    async def run() -> None:
        with mini_racer() as mr:
            async with mr.wrap_py_function(append) as jsfunc:
                # "Install" our JS function on the global "this" object:
                cast("JSFunction", mr.eval("x => this.func = x"))(jsfunc)

                pending = [
                    cast("JSPromise", mr.eval("this.func(42)")) for _ in range(100)
                ]

                assert await gather(*pending) == ["foobar"] * 100

            assert data == [(42,)] * 100
            data.clear()
            del pending, jsfunc
            await async_assert_no_v8_objects(mr)

    data.clear()

    start = time()
    asyncio_run(run())
    # The above should run in just over a second.
    # Just verify it didn't take 100 seconds (i.e., that things didn't execute
    # sequentially):
    assert time() - start < 10  # noqa: PLR2004


def test_call_on_exit() -> None:
    """Checks that calls from JS made while we're trying to tear down the wrapped
    function are ignored and don't break anything."""

    data = []

    async def run() -> None:
        append_called_fut: Future[None] = Future()

        async def append(*args: Any) -> str:  # noqa: ANN401
            data.append(args)
            append_called_fut.set_result(None)
            # sleep long enough that the test will fail unless this is either
            # interrupted, or never started to begin with:
            await asyncio_sleep(10000)
            return "foobar"

        with mini_racer() as mr:
            async with mr.wrap_py_function(append) as jsfunc:
                # "Install" our JS function on the global "this" object:
                cast("JSFunction", mr.eval("x => this.func = x"))(jsfunc)

                # Note: we don't await the promise, meaning we just start a call and
                # never finish it:
                assert isinstance(mr.eval("this.func(42)"), JSPromise)

                await append_called_fut

                # After this line, we start tearing down the mr.wrap_py_function context
                # manager, which entails stopping the call processor.
                # Let's make sure we don't fall over ourselves (it's fair to either
                # process the last straggling calls, or ignore them, but make sure we
                # don't hang).

        del jsfunc
        await async_assert_no_v8_objects(mr)

    for _ in range(_NUM_LOOPS):
        data[:] = []
        asyncio_run(run())



"""Python wrappers for JavaScript object types."""

from __future__ import annotations

import asyncio
import ctypes
from datetime import datetime, timezone
from operator import index as op_index
from typing import TYPE_CHECKING, Any, ClassVar, cast

from py_mini_racer._exc import (
    JSArrayIndexError,
    JSConversionException,
    JSEvalException,
    JSKeyError,
    JSOOMException,
    JSParseException,
    JSTerminatedException,
    JSTimeoutException,
    JSValueError,
)
from py_mini_racer._types import (
    CancelableJSFunction,
    JSArray,
    JSFunction,
    JSMappedObject,
    JSObject,
    JSPromise,
    JSSymbol,
    JSUndefined,
    JSUndefinedType,
    PythonJSConvertedTypes,
)

if TYPE_CHECKING:
    from collections.abc import Generator, Iterator, Sequence

    from py_mini_racer._context import Context
    from py_mini_racer._dll import RawValueHandleTypeImpl
    from py_mini_racer._value_handle import ValueHandle


class JSObjectImpl(JSObject):
    def __init__(self, ctx: Context, handle: ValueHandle) -> None:
        self._ctx = ctx
        self._handle = handle

    def __hash__(self) -> int:
        return self._ctx.get_identity_hash(self)

    @property
    def raw_handle(self) -> ValueHandle:
        return self._handle


class JSMappedObjectImpl(JSObjectImpl, JSMappedObject):
    def __iter__(self) -> Iterator[PythonJSConvertedTypes]:
        return iter(self._get_own_property_names())

    def __getitem__(self, key: PythonJSConvertedTypes) -> PythonJSConvertedTypes:
        return self._ctx.get_object_item(self, key)

    def __setitem__(
        self, key: PythonJSConvertedTypes, val: PythonJSConvertedTypes
    ) -> None:
        self._ctx.set_object_item(self, key, val)

    def __delitem__(self, key: PythonJSConvertedTypes) -> None:
        self._ctx.del_object_item(self, key)

    def __len__(self) -> int:
        return len(self._get_own_property_names())

    def _get_own_property_names(self) -> tuple[PythonJSConvertedTypes, ...]:
        return self._ctx.get_own_property_names(self)


class JSArrayImpl(JSArray, JSObjectImpl):
    def __len__(self) -> int:
        return cast("int", self._ctx.get_object_item(self, "length"))

    def __getitem__(self, index: int | slice) -> Any:  # noqa: ANN401
        if not isinstance(index, int):
            raise TypeError

        index = op_index(index)
        if index < 0:
            index += len(self)

        if 0 <= index < len(self):
            return self._ctx.get_object_item(self, index)

        raise IndexError

    def __setitem__(self, index: int | slice, val: Any) -> None:  # noqa: ANN401
        if not isinstance(index, int):
            raise TypeError

        self._ctx.set_object_item(self, index, val)

    def __delitem__(self, index: int | slice) -> None:
        if not isinstance(index, int):
            raise TypeError

        if index >= len(self) or index < -len(self):
            # JavaScript Array.prototype.splice() just ignores deletion beyond the
            # end of the array, meaning if you pass a very large value here it would
            # do nothing. Likewise, it just caps negative values at the length of the
            # array, meaning if you pass a very negative value here it would just
            # delete element 0.
            # For consistency with Python lists, let's tell the caller they're out of
            # bounds:
            raise JSArrayIndexError

        self._ctx.del_from_array(self, index)

    def insert(self, index: int, new_obj: PythonJSConvertedTypes) -> None:
        self._ctx.array_insert(self, index, new_obj)

    def __iter__(self) -> Iterator[PythonJSConvertedTypes]:
        for i in range(len(self)):
            yield self._ctx.get_object_item(self, i)

    def append(self, value: PythonJSConvertedTypes) -> None:
        self._ctx.array_push(self, value)


class JSFunctionImpl(JSMappedObjectImpl, JSFunction):
    def __call__(
        self,
        *args: PythonJSConvertedTypes,
        this: JSObject | JSUndefinedType = JSUndefined,
        timeout_sec: float | None = None,
    ) -> PythonJSConvertedTypes:
        if not self._ctx.are_we_running_on_the_mini_racer_event_loop():

            async def run() -> PythonJSConvertedTypes:
                try:
                    return await asyncio.wait_for(
                        self._ctx.call_function_cancelable(self, *args, this=this),
                        timeout=timeout_sec,
                    )
                except asyncio.TimeoutError as e:
                    raise JSTimeoutException from e

            return asyncio.run_coroutine_threadsafe(
                run(), self._ctx.event_loop
            ).result()

        assert timeout_sec is None, (
            "To apply a timeout in an async context, use "
            "`await asyncio.wait_for(your_func.cancelable()(your_params), "
            "timeout=your_timeout)`"
        )

        return self._ctx.call_function(self, *args, this=this)

    def cancelable(self) -> CancelableJSFunction:
        return CancelableJSFunctionImpl(self._ctx, self._handle)


class CancelableJSFunctionImpl(JSMappedObjectImpl, CancelableJSFunction):
    async def __call__(
        self,
        *args: PythonJSConvertedTypes,
        this: JSObject | JSUndefinedType = JSUndefined,
    ) -> PythonJSConvertedTypes:
        return await self._ctx.call_function_cancelable(self, *args, this=this)


class JSSymbolImpl(JSMappedObjectImpl, JSSymbol):
    pass


class JSPromiseImpl(JSObjectImpl, JSPromise):
    def get(self, *, timeout: float | None = None) -> PythonJSConvertedTypes:
        assert not self._ctx.are_we_running_on_the_mini_racer_event_loop(), (
            "In an async context, call `await promise` instead of promise.get()"
        )

        async def run() -> PythonJSConvertedTypes:
            try:
                return await asyncio.wait_for(
                    self._ctx.await_promise(self), timeout=timeout
                )
            except asyncio.TimeoutError as e:
                raise JSTimeoutException from e

        return asyncio.run_coroutine_threadsafe(run(), self._ctx.event_loop).result()

    def __await__(self) -> Generator[Any, None, Any]:
        return self._ctx.await_promise(self).__await__()


class _ArrayBufferByte(ctypes.Structure):
    # Cannot use c_ubyte directly because it uses <B
    # as an internal type but we need B for memoryview.
    _fields_: ClassVar[Sequence[tuple[str, type]]] = [("b", ctypes.c_ubyte)]
    _pack_ = 1


class _MiniRacerTypes:
    """MiniRacer types identifier

    Note: it needs to be coherent with mini_racer.cc.
    """

    invalid = 0
    null = 1
    bool = 2
    integer = 3
    double = 4
    str_utf8 = 5
    array = 6
    # deprecated:
    hash = 7
    date = 8
    symbol = 9
    object = 10
    undefined = 11

    function = 100
    shared_array_buffer = 101
    array_buffer = 102
    promise = 103

    execute_exception = 200
    parse_exception = 201
    oom_exception = 202
    timeout_exception = 203
    terminated_exception = 204
    value_exception = 205
    key_exception = 206


_ERRORS: dict[int, tuple[type[JSEvalException], str]] = {
    _MiniRacerTypes.parse_exception: (
        JSParseException,
        "Unknown JavaScript error during parse",
    ),
    _MiniRacerTypes.execute_exception: (
        JSEvalException,
        "Uknown JavaScript error during execution",
    ),
    _MiniRacerTypes.oom_exception: (JSOOMException, "JavaScript memory limit reached"),
    _MiniRacerTypes.terminated_exception: (
        JSTerminatedException,
        "JavaScript was terminated",
    ),
    _MiniRacerTypes.key_exception: (JSKeyError, "No such key found in object"),
    _MiniRacerTypes.value_exception: (
        JSValueError,
        "Bad value passed to JavaScript engine",
    ),
}


class ObjectFactoryImpl:
    def value_handle_to_python(  # noqa: C901, PLR0911, PLR0912
        self, ctx: Context, val_handle: ValueHandle
    ) -> PythonJSConvertedTypes:
        """Convert a value handle from the C++ side into a Python object."""

        # A MiniRacer value handle is a pointer to a structure which, for some
        # simple types like ints, floats, and strings, is sufficient to describe the
        # data, enabling us to convert the value immediately and free the handle.

        # For more complex types, like Objects and Arrays, the handle is just an opaque
        # pointer to a V8 object. In these cases, we retain the value handle,
        # wrapping it in a Python object. We can then use the handle in follow-on API
        # calls to work with the underlying V8 object.

        # In either case the handle is owned by the C++ side. It's the responsibility
        # of the Python side to call mr_free_value() when done with with the handle
        # to free up memory, but the C++ side will eventually free it on context
        # teardown either way.

        raw = cast("RawValueHandleTypeImpl", val_handle.raw)

        typ = raw.contents.type
        val = raw.contents.value
        length = raw.contents.len

        error_info = _ERRORS.get(raw.contents.type)
        if error_info:
            klass, generic_msg = error_info

            msg = val.bytes_val[0:length].decode("utf-8") or generic_msg
            raise klass(msg)

        if typ == _MiniRacerTypes.null:
            return None
        if typ == _MiniRacerTypes.undefined:
            return JSUndefined
        if typ == _MiniRacerTypes.bool:
            return bool(val.int_val == 1)
        if typ == _MiniRacerTypes.integer:
            return int(val.int_val)
        if typ == _MiniRacerTypes.double:
            return float(val.double_val)
        if typ == _MiniRacerTypes.str_utf8:
            return str(val.bytes_val[0:length].decode("utf-8"))
        if typ == _MiniRacerTypes.function:
            return JSFunctionImpl(ctx, val_handle)
        if typ == _MiniRacerTypes.date:
            timestamp = val.double_val
            # JS timestamps are milliseconds. In Python we are in seconds:
            return datetime.fromtimestamp(timestamp / 1000.0, timezone.utc)
        if typ == _MiniRacerTypes.symbol:
            return JSSymbolImpl(ctx, val_handle)
        if typ in (_MiniRacerTypes.shared_array_buffer, _MiniRacerTypes.array_buffer):
            buf = _ArrayBufferByte * length
            cdata = buf.from_address(val.value_ptr)
            # Save a reference to the context to prevent garbage collection of the
            # backing store:
            cdata._origin = ctx  # noqa: SLF001
            result = memoryview(cdata)
            # Avoids "NotImplementedError: memoryview: unsupported format T{<B:b:}"
            # in Python 3.12:
            return result.cast("B")

        if typ == _MiniRacerTypes.promise:
            return JSPromiseImpl(ctx, val_handle)

        if typ == _MiniRacerTypes.array:
            return JSArrayImpl(ctx, val_handle)

        if typ == _MiniRacerTypes.object:
            return JSMappedObjectImpl(ctx, val_handle)

        raise JSConversionException

    def python_to_value_handle(  # noqa: PLR0911
        self, ctx: Context, obj: PythonJSConvertedTypes
    ) -> ValueHandle:
        if isinstance(obj, JSObjectImpl):
            # JSObjects originate from the V8 side. We can just send back the handle
            # we originally got. (This also covers derived types JSFunction, JSSymbol,
            # JSPromise, and JSArray.)
            return obj.raw_handle

        if obj is None:
            return ctx.create_intish_val(0, _MiniRacerTypes.null)
        if obj is JSUndefined:
            return ctx.create_intish_val(0, _MiniRacerTypes.undefined)
        if isinstance(obj, bool):
            return ctx.create_intish_val(1 if obj else 0, _MiniRacerTypes.bool)
        if isinstance(obj, int):
            if obj - 2**31 <= obj < 2**31:
                return ctx.create_intish_val(obj, _MiniRacerTypes.integer)

            # We transmit ints as int32, so "upgrade" to double upon overflow.
            # (ECMAScript numeric is double anyway, but V8 does internally distinguish
            # int types, so we try and preserve integer-ness for round-tripping
            # purposes.)
            # JS BigInt would be a closer representation of Python int, but upgrading
            # to BigInt would probably be surprising for most applications, so for now,
            # we approximate with double:
            return ctx.create_doublish_val(obj, _MiniRacerTypes.double)
        if isinstance(obj, float):
            return ctx.create_doublish_val(obj, _MiniRacerTypes.double)
        if isinstance(obj, str):
            return ctx.create_string_val(obj, _MiniRacerTypes.str_utf8)
        if isinstance(obj, datetime):
            # JS timestamps are milliseconds. In Python we are in seconds:
            return ctx.create_doublish_val(
                obj.timestamp() * 1000.0, _MiniRacerTypes.date
            )

        # Note: we skip shared array buffers, so for now at least, handles to shared
        # array buffers can only be transmitted from JS to Python.

        raise JSConversionException
